from decimal import Decimal

class Encrypted(dict):
    def __init__(self, ciphertext: Decimal, offsetCount: Decimal):
        ''' All values are read only '''
        self._ciphertext = ciphertext
        self._offsetCount = offsetCount

    def __str__(self):
        return f"quarkz.Encryption object @ {self._offsetCount}"
    
    @property
    def __dict__(self):
        return {"ciphertext": self._ciphertext, "offset": self._offsetCount}


class KeyPair():
    def __init__(self, public_key: dict, private_key: dict) -> None: 
        self._pub = public_key
        self.__priv = private_key #private to class

    @property
    def __dict__(self):
        return {"publicKey": self._pub, "privateKey": self.__priv}

    def get_public_key(self):
        return self._pub

    def get_private_key(self):
        return self.__priv





from Crypto.Util import number
from decimal import Decimal
import decimal
from quarkz.dtypes import KeyPair
import random
import sys
import binascii


decimal.getcontext().prec=100000

def gcd(a, b): 
   while a != 0:
      a, b = b % a, a
   return b

def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    g, y, x = egcd(b%a,a)
    return (g, x - (b//a) * y, y)

def mod_inverse(a, m):
    g, x, y = egcd(a, m)
    if g != 1:
        raise Exception('No modular inverse')
    return x%m

def modpow(x,n,m):
  if n == 0:
    return 1
  elif n == 1:
    return x
  elif n%2 == 0:
    return modpow(x*(x%m),n/2,m)%m
  elif n%2 == 1:
    return (x *  modpow(x*(x%m),(n-1)/2,m)%m )%m

def convert_to_int(data: str) -> int: 
    return int.from_bytes(data.encode(), 'big')

def convert_to_str(data: int) -> str:
    return data.to_bytes((data.bit_length() + 7) // 8, 'big').decode()

    



def createKey(keysize: int = 1024):
    p = number.getPrime(keysize)
    q = number.getPrime(keysize)
    n = Decimal(p*q)
    #print(n)
    phi = Decimal((p-1)*(q-1))
    while True:
        e = Decimal(number.getPrime(8))
        r = gcd(int(e), int(phi))
        if r == 1:
            t = Decimal(random.getrandbits(4))
            o = e**t
            break

    d = Decimal(mod_inverse(int(e), int(phi)))

    #print ("osize: ", sys.getsizeof(o)*8)
    
    diff = (abs(n-Decimal(o))) % n

    #print(n)

    x = Decimal(random.getrandbits(8192))

    if diff > 0:
        ratio = (n/diff) * x
        #print ("ratio: ", sys.getsizeof(round(ratio))*8)
    else:
        u = random.randint(0, 1000)
        o -= u
        diff = abs(n-o) % n
        ratio = (n/diff) * x

    privateKey = {
        "d": d,
        "n": n,
        "diff": diff,
        "r": (n/diff),
    }

    publicKey = {
        "e": e,
        "o": o,
        "ratio": ratio,
        "diff": diff,
        "x": x,
        "n": n
    }

    keyPair = {
        "private_key": privateKey,
        "public_key": publicKey
    }

    return KeyPair(**keyPair)



import random
import sys
import json
import decimal

from base64 import b64encode, b64decode
from decimal import Decimal

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util import number
from Crypto.PublicKey import RSA 

from quarkz import utils 
from quarkz.dtypes import Encrypted 
import quarkz

decimal.getcontext().prec=100000


def encrypt(message: str, publicKey: dict) -> quarkz.dtypes.Encrypted: 

    assert(type(message) == str)

    message = utils.convert_to_int(message)

    m = Decimal(message)

    # This is really slow find ways to speed up...
    s = m**publicKey["e"]
    
    #print ("s: ", sys.getsizeof(s)*8)


    # Calculating the count is really slow
    # We want to implement some sort of 
    # modular exponentiation trick to speed this up.
    count = Decimal(int(s) // int(publicKey["o"]))

    #print("count: ", sys.getsizeof(count) * 8)

    offsetCount = Decimal(count) % Decimal(publicKey["ratio"])

    n1 = (publicKey["ratio"] * publicKey["o"]) / (publicKey["ratio"] - 1)

    d = n1 / publicKey["ratio"]

    n2 = d + publicKey["o"]

    #print ("n: ", publicKey["n"])

    # for x in range (10000):

    #     #guess = random.getrandbits(8192)
    #     guess = number.getPrime(257) * number.getPrime(257)

    #     #pred = abs((publicKey["o"] * publicKey["ratio"]) / (guess - publicKey["ratio"]))
    #     pred = abs(publicKey["ratio"] - ((publicKey["o"] * publicKey["ratio"]) / guess))
    #     #print (pred)
        

    #print (round(n1))
    #print (round(n2))

    #print (round(publicKey["ratio"]))

    #print("offset: ", sys.getsizeof(round(offsetCount))*8)

    ciphertext = Decimal(pow(m, publicKey["e"], publicKey["o"]))

    #print ("ciphertext generated: ", ciphertext)

    data = {"ciphertext": ciphertext, "offsetCount": offsetCount}

    return Encrypted(**data)


def decrypt(encrypted: quarkz.dtypes.Encrypted, keypair: quarkz.dtypes.KeyPair) -> int:
    encrypted = vars(encrypted)

    privateKey = keypair.get_private_key()

    offset = round(((encrypted["offset"]) * privateKey["diff"])) % privateKey["n"]

    ciphertext = int(encrypted["ciphertext"] - offset)

    plaintext = pow(ciphertext, int(privateKey["d"]), int(privateKey["n"]))

    #print (plaintext)
    
    if plaintext:
        return utils.convert_to_str(plaintext)
    else: 
        ciphertext = int(encrypted["ciphertext"] + offset)
        return pow(ciphertext, int(privateKey["d"]), int(privateKey["n"]))



if __name__ == "__main__":
    #first, create a new key pair 
    pair = utils.createKey(1024)

    #encrypt some data
    message = 69
    public_key = pair.get_public_key()
    encrypted_data = encrypt(message, public_key)

    #decrypt the data again
    decrypted_data = decrypt(encrypted_data, pair)
    print(decrypted_data)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

setup(
  name='quarkz',
  version='0.0.1',
  url='https://github.com/quarkz-encryption/Quarkz',
  author='tavivia',
  author_email='tavian@quarkz-encryption.com',
  packages=find_packages(exclude=("tests",)),
  platforms='any',
  license='MIT',
  ext_modules=[],
  description="An Encryption library",
  long_description='See https://github.com/quarkz-encryption/quarkz',
)
